@@ -1,57 +1,33 @@
 ;;; installed-scm-file
 
-;;;; 	Copyright (C) 2000, 2001 Free Software Foundation, Inc.
+;;;; Copyright (C) 2000,2001,2002, 2006, 2009, 2010, 2013, 2015 Free Software Foundation, Inc.
 ;;;; 
-;;;; This program is free software; you can redistribute it and/or modify
-;;;; it under the terms of the GNU General Public License as published by
-;;;; the Free Software Foundation; either version 2, or (at your option)
-;;;; any later version.
+;;;; This library is free software; you can redistribute it and/or
+;;;; modify it under the terms of the GNU Lesser General Public
+;;;; License as published by the Free Software Foundation; either
+;;;; version 3 of the License, or (at your option) any later version.
 ;;;; 
-;;;; This program is distributed in the hope that it will be useful,
+;;;; This library is distributed in the hope that it will be useful,
 ;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
-;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-;;;; GNU General Public License for more details.
+;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+;;;; Lesser General Public License for more details.
 ;;;; 
-;;;; You should have received a copy of the GNU General Public License
-;;;; along with this software; see the file COPYING.  If not, write to
-;;;; the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
-;;;; Boston, MA 02111-1307 USA
-;;;;
-;;;; As a special exception, the Free Software Foundation gives permission
-;;;; for additional uses of the text contained in its release of GUILE.
-;;;;
-;;;; The exception is that, if you link the GUILE library with other files
-;;;; to produce an executable, this does not by itself cause the
-;;;; resulting executable to be covered by the GNU General Public License.
-;;;; Your use of that executable is in no way restricted on account of
-;;;; linking the GUILE library code into it.
-;;;;
-;;;; This exception does not however invalidate any other reasons why
-;;;; the executable file might be covered by the GNU General Public License.
-;;;;
-;;;; This exception applies only to the code released by the
-;;;; Free Software Foundation under the name GUILE.  If you copy
-;;;; code from other Free Software Foundation releases into a copy of
-;;;; GUILE, as the General Public License permits, the exception does
-;;;; not apply to the code that you add in this way.  To avoid misleading
-;;;; anyone as to the status of such modified files, you must delete
-;;;; this exception notice from them.
-;;;;
-;;;; If you write modifications of your own for GUILE, it is your choice
-;;;; whether to permit this exception to apply to your modifications.
-;;;; If you do not wish that, delete this exception notice.
+;;;; You should have received a copy of the GNU Lesser General Public
+;;;; License along with this library; if not, write to the Free Software
+;;;; Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 ;;;; 
 
 
 (define-module (oop goops save)
   :use-module (oop goops internal)
-  :use-module (oop goops util)
-  :re-export (make-unbound)
-  :export (save-objects load-objects restore
+  :export (make-unbound save-objects load-objects restore
 	   enumerate! enumerate-component!
 	   write-readably write-component write-component-procedure
 	   literal? readable make-readable))
 
+(define (make-unbound)
+  *unbound*)
+
 ;;;
 ;;; save-objects ALIST PORT [EXCLUDED] [USES]
 ;;;
@@ -135,14 +111,10 @@
 ;;; Readables
 ;;;
 
-(if (or (not (defined? 'readables))
-	(not readables))
-    (define readables (make-weak-key-hash-table 61)))
+(define readables (make-weak-key-hash-table 61))
 
-(define readable
-  (procedure->memoizing-macro
-    (lambda (exp env)
-      `(make-readable ,(cadr exp) ',(copy-tree (cadr exp))))))
+(define-macro (readable exp)
+  `(make-readable ,exp ',(copy-tree exp)))
 
 (define (make-readable obj expr)
   (hashq-set! readables obj expr)
@@ -151,8 +123,33 @@
 (define (readable-expression obj)
   `(readable ,(hashq-ref readables obj)))
 
+;; FIXME: if obj is nil or false, this can return a false value.  OTOH
+;; usually this is only for non-immediates.
 (define (readable? obj)
-  (hashq-get-handle readables obj))
+  (hashq-ref readables obj))
+
+;;;
+;;; Writer helpers
+;;;
+
+(define (write-component-procedure o file env)
+  "Return #f if circular reference"
+  (cond ((immediate? o) (write o file) #t)
+	((readable? o) (write (readable-expression o) file) #t)
+	((excluded? o env) (display #f file) #t)
+	(else
+	 (let ((info (object-info o env)))
+	   (cond ((not (binding? info)) (write-readably o file env) #t)
+		 ((not (eq? (visiting info) #:defined)) #f) ;forward reference
+		 (else (display (binding info) file) #t))))))
+
+;;; write-component OBJECT PATCHER FILE ENV
+;;;
+(define-macro (write-component object patcher file env)
+  `(or (write-component-procedure ,object ,file ,env)
+       (begin
+         (display #f ,file)
+         (add-patcher! ,patcher ,env))))
 
 ;;;
 ;;; Strings
@@ -212,8 +209,7 @@
   (enumerate-component! (shared-array-root o) env))
 
 (define (make-mapper array)
-  (let* ((dims (array-dimensions array))
-	 (n (array-rank array))
+  (let* ((n (array-rank array))
 	 (indices (reverse (if (<= n 11)
 			       (list-tail '(t s r q p n m l k j i)  (- 11 n))
 			       (let loop ((n n)
@@ -300,7 +296,7 @@
 	   (display "(list->uniform-array " file)
 	   (display (array-rank o) file)
 	   (display " '() " file)
-	   (write-array "(list " o file env)))))
+	   (write-array "(list " o #f file env)))))
 
 ;;;
 ;;; Pairs
@@ -374,44 +370,41 @@
 ;; Don't export this function!  This is all very temporary.
 ;;
 (define (get-set-for-each proc class)
-  (for-each (lambda (slotdef g-n-s)
-	      (let ((g-n-s (cddr g-n-s)))
-		(cond ((integer? g-n-s)
-		       (proc (standard-get g-n-s) (standard-set g-n-s)))
-		      ((not (memq (slot-definition-allocation slotdef)
-				  '(#:class #:each-subclass)))
-		       (proc (car g-n-s) (cadr g-n-s))))))
-	    (class-slots class)
-	    (slot-ref class 'getters-n-setters)))
+  (for-each (lambda (slot)
+	      (unless (memq (slot-definition-allocation slot)
+                            '(#:class #:each-subclass))
+                (let ((ref (slot-definition-slot-ref slot))
+                      (set (slot-definition-slot-set! slot))
+                      (index (slot-definition-index slot)))
+                  (if ref
+                      (proc ref set)
+                      (proc (standard-get index) (standard-set index))))))
+	    (class-slots class)))
 
 (define (access-for-each proc class)
-  (for-each (lambda (slotdef g-n-s)
-	      (let ((g-n-s (cddr g-n-s))
-		    (a (slot-definition-accessor slotdef)))
-		(cond ((integer? g-n-s)
-		       (proc (slot-definition-name slotdef)
-			     (and a (generic-function-name a))
-			     (standard-get g-n-s)
-			     (standard-set g-n-s)))
-		      ((not (memq (slot-definition-allocation slotdef)
-				  '(#:class #:each-subclass)))
-		       (proc (slot-definition-name slotdef)
-			     (and a (generic-function-name a))
-			     (car g-n-s)
-			     (cadr g-n-s))))))
-	    (class-slots class)
-	    (slot-ref class 'getters-n-setters)))
+  (for-each (lambda (slot)
+              (unless (memq (slot-definition-allocation slot)
+                            '(#:class #:each-subclass))
+                (let ((name (slot-definition-name slot))
+                      (accessor (and=> (slot-definition-accessor slot)
+                                       generic-function-name))
+                      (ref (slot-definition-slot-ref slot))
+                      (set (slot-definition-slot-set! slot))
+                      (index (slot-definition-index slot)))
+                  (if ref
+                      (proc name accessor ref set)
+                      (proc name accessor
+                            (standard-get index) (standard-set index))))))
+	    (class-slots class)))
 
-(define restore
-  (procedure->macro
-    (lambda (exp env)
-      "(restore CLASS (SLOT-NAME1 ...) EXP1 ...)"
-      `(let ((o (,%allocate-instance ,(cadr exp) '())))
-	 (for-each (lambda (name val)
-		     (,slot-set! o name val))
-		   ',(caddr exp)
-		   (list ,@(cdddr exp)))
-	 o))))
+(define-macro (restore class slots . exps)
+  "(restore CLASS (SLOT-NAME1 ...) EXP1 ...)"
+  `(let ((o ((@@ (oop goops) %allocate-instance) ,class '())))
+     (for-each (lambda (name val)
+                 (slot-set! o name val))
+               ',slots
+               (list ,@exps))
+     o))
 
 (define-method (enumerate! (o <object>) env)
   (get-set-for-each (lambda (get set)
@@ -633,26 +626,6 @@
 	   (pop-ref! env)
 	   (set! (objects env) (cons o (objects env)))))))
 
-(define (write-component-procedure o file env)
-  "Return #f if circular reference"
-  (cond ((immediate? o) (write o file) #t)
-	((readable? o) (write (readable-expression o) file) #t)
-	((excluded? o env) (display #f file) #t)
-	(else
-	 (let ((info (object-info o env)))
-	   (cond ((not (binding? info)) (write-readably o file env) #t)
-		 ((not (eq? (visiting info) #:defined)) #f) ;forward reference
-		 (else (display (binding info) file) #t))))))
-
-;;; write-component OBJECT PATCHER FILE ENV
-;;;
-(define write-component
-  (procedure->memoizing-macro
-    (lambda (exp env)
-      `(or (write-component-procedure ,(cadr exp) ,@(cdddr exp))
-	   (begin
-	     (display #f ,(cadddr exp))
-	     (add-patcher! ,(caddr exp) env))))))
 
 ;;;
 ;;; Main engine
@@ -882,10 +855,12 @@
     (close-port port)
     objects))
 
+(define iface (module-public-interface (current-module)))
+
 (define-method (load-objects (file <input-port>))
   (let ((m (make-module)))
     (module-use! m the-scm-module)
-    (module-use! m %module-public-interface)
+    (module-use! m iface)
     (save-module-excursion
      (lambda ()
        (set-current-module m)
